#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <math.h>
#include <algorithm>

#include <wchar.h>
#include <windows.h>
#include <fcntl.h>
#define	_O_U16TEXT 0x00020000
#define	_O_U8TEXT 0x00040000


int	result =EXIT_SUCCESS;

void	loadAndProcessFile (wchar_t* fileName) {
	FILE *handle =_wfopen(fileName, L"rb");
	if (handle) {
	std::vector<uint8_t> header;
	while (!feof(handle) && header.size() <16) header.push_back(fgetc(handle));
	if (header.size() ==16 && memcmp(&header[0], "NES\x1A", 4) ==0) {
		// iNES PRG-ROM size
		size_t prgBytes =(header[4]? header[4]: 256) *16384;
		
		if ((header[7] &0xC) ==0x8) { // NES 2.0 PRG-ROM size
			if ((header[9] &0xF) !=0xF) {
				int prgBanks =((header[9] &0xF) <<8) | header[4];
				prgBytes =(prgBanks? prgBanks: 4096) *16384;
			} else
				prgBytes =pow(2, header[4] >>2) *((header[4] &3)*2 +1);
		}
		std::vector<uint8_t> prgROM;
		std::vector<uint8_t>::iterator pos;
		while (!feof(handle) && prgROM.size() <prgBytes) prgROM.push_back(fgetc(handle));
		
		unsigned char ldy4016 [3] = { 0xAC, 0x16, 0x40 };
		unsigned char lda4016 [3] = { 0xAD, 0x16, 0x40 };
		unsigned char ldx4016 [3] = { 0xAE, 0x16, 0x40 };
		unsigned char lda4016x[3] = { 0xBD, 0x16, 0x40 };
		unsigned char lda4016y[3] = { 0xB9, 0x16, 0x40 };
		unsigned char and04   [2] = { 0x29, 0x04 };
		
		std::vector<uint32_t> foundPositions;
		pos =prgROM.begin();
		while ((pos =std::search(pos +1, prgROM.end(), ldy4016, ldy4016 +3)) !=prgROM.end()) if (std::search(pos, pos +32, and04, and04 +2) !=pos +32) foundPositions.push_back(pos -prgROM.begin());
		pos =prgROM.begin();
		while ((pos =std::search(pos +1, prgROM.end(), lda4016, lda4016 +3)) !=prgROM.end()) if (std::search(pos, pos +32, and04, and04 +2) !=pos +32) foundPositions.push_back(pos -prgROM.begin());
		pos =prgROM.begin();
		while ((pos =std::search(pos +1, prgROM.end(), ldx4016, ldx4016 +3)) !=prgROM.end()) if (std::search(pos, pos +32, and04, and04 +2) !=pos +32) foundPositions.push_back(pos -prgROM.begin());
		pos =prgROM.begin();
		while ((pos =std::search(pos +1, prgROM.end(), lda4016x, lda4016x +3)) !=prgROM.end()) if (std::search(pos, pos +32, and04, and04 +2) !=pos +32) foundPositions.push_back(pos -prgROM.begin());
		pos =prgROM.begin();
		while ((pos =std::search(pos +1, prgROM.end(), lda4016y, lda4016y +3)) !=prgROM.end()) if (std::search(pos, pos +32, and04, and04 +2) !=pos +32) foundPositions.push_back(pos -prgROM.begin());
		
		if (foundPositions.size()) {
			wprintf(L"%s:", fileName);
			for (auto& foundPosition: foundPositions) wprintf(L" %05X", foundPosition);
			wprintf(L"\n");
		}
		
		fclose(handle);
	} else {
		fwprintf(stderr, L"%s: Not a NES 2.0 format file\n", fileName);
		result =EXIT_FAILURE;
	}
	
	fclose(handle);
	} else {
		_wperror(fileName);
		result =EXIT_FAILURE;
	}
}

int	main (int argc, char **argv) {
	int	nArgs;
	LPWSTR *szArgList = CommandLineToArgvW(GetCommandLineW(), &nArgs);

	_setmode(_fileno(stdout), _O_U8TEXT);
	_setmode(_fileno(stderr), _O_U8TEXT);
	for (int arg =1; arg <nArgs; arg++) {
		WIN32_FIND_DATAW result;
		HANDLE hFind =FindFirstFileW(szArgList[arg], &result);
		if (hFind ==INVALID_HANDLE_VALUE) {
			fwprintf(stderr, L"%s: No matching files found\n", szArgList[arg]);
			break;
		}
		do {
			loadAndProcessFile(result.cFileName);
		} while (FindNextFileW(hFind, &result));
		FindClose(hFind);
	}
	return result;
}